# --------------------------------------------------------------------
# Copyright © 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

import functools
import json
import subprocess

import apt_pkg

from click.json_helpers import (
    json_array_to_python,
    json_object_to_python,
)
from click.repository import (
    get_repository,
    ClickRepositoryError,
)
from gi.repository import Click

from . import datasource
from .utils import sudo


def is_framework(manifest):
    """Return True if the given click is a framework"""
    for app in manifest.get("hooks", {}):
        for hook in manifest["hooks"][app]:
            if hook == "framework":
                return True
    return False


class ClickDataSource(datasource.DataSource):

    @property
    def installed_frameworks(self):
        return [f.get_name().rsplit("-", 2)[0]
                for f in Click.Framework.get_frameworks()
                # we always hide the ubuntu-core-15.04 etc frameworks
                if not f.get_name().startswith("ubuntu-")
                ]

    @property
    def installed_apps(self):
        # FIXME: provide --user
        return [app["name"]
                for app in self._get_clicks_for_user()
                if not is_framework(app)]

    def details(self, part):
        repo = get_repository()
        try:
            manifest = repo.details(part)
            return manifest
        except ClickRepositoryError:
            return None

    def search(self, terms):
        repo = get_repository()
        try:
            # FIXME: how does the store expects a AND search(?)
            results = repo.search(",".join(terms))
            return results
        except ClickRepositoryError:
            return None

    @property
    def upgradable_apps(self):
        # FIXME: provide --user
        repo = get_repository()
        all_updates = []
        all_updates_list = repo.get_upgradable()
        for current, new in all_updates_list:
            all_updates.append(new["name"])
        return all_updates

    def versions(self, all_versions=False):
        # FIXME: provide --user
        repo = get_repository()
        try:
            all_updates_list = repo.get_upgradable()
            all_updates = {}
            for current, new in all_updates_list:
                all_updates[current["name"]] = new
        except ClickRepositoryError:
            return []
        parts = []
        user_manifests = self._get_clicks_for_user()
        if all_versions:
            manifests = self._get_all_clicks()
        else:
            manifests = self._get_clicks_for_user()
        for manifest in manifests:
            if manifest["name"] in all_updates:
                latest_version = all_updates[manifest["name"]]["version"]
            else:
                latest_version = ""
            parts.append(datasource.Part(
                manifest["name"],
                manifest.get("current_stability", "edge"),
                manifest["version"],
                latest_version,
                manifest.get("checksum-sha512", "no-hash"),
                # FIXME: ugly
                manifest["name"] +
                manifest["version"]
                in [m["name"] + m["version"] for m in user_manifests],
            ))
        return parts

    def update(self, names_to_update=None, list_only=False):
        cmd = ["click", "update", '--meter']
        if names_to_update:
            cmd += names_to_update
        if list_only:
            cmd += "--list"
        # FIXME: make this a config option
        cmd.append('--all-users')
        subprocess.call(cmd)

    def rollback(self, names_to_rollback=None):
        if names_to_rollback is None:
            return True
        sudo()
        db = Click.DB()
        db.read()
        registry = Click.User.for_all_users(db)
        installed = db.get_packages(all_versions=True)
        for name in names_to_rollback:
            old_version = registry.get_version(name)
            rollback_targets = [pkg.props.version
                                for pkg in installed
                                if pkg.props.package == name and
                                pkg.props.version != old_version]
            if len(rollback_targets) < 1:
                print("Can not find a rollback version")
                return False
            rollback_version = sorted(
                rollback_targets,
                key=functools.cmp_to_key(apt_pkg.version_compare))[-1]
            print("Rolling back {} to version {}".format(
                name, rollback_version))
            registry.set_version(name, rollback_version)
            # FIXME: do we want this?
            db.maybe_remove(name, old_version)
        return True

    def _get_all_clicks(self):
        db = Click.DB()
        db.read(db_dir=None)
        manifests = []
        for pkg in db.get_packages(True):
            manifest = db.get_manifest(
                pkg.get_property("package"), pkg.get_property("version"))
            manifests.append(json_object_to_python(manifest))
        return manifests

    def _get_clicks_for_user(self, user=None):
        db = Click.DB()
        db.read(db_dir=None)
        registry = Click.User.for_user(db)
        return json_array_to_python(registry.get_manifests())

    def get_tags(self, part):
        try:
            json_str = subprocess.check_output(
                ["click", "info", part],
                universal_newlines=True)
        except subprocess.CalledProcessError:
            return []
        if not json_str:
            return []
        manifest = json.loads(json_str)
        stability = manifest.get("available-stability-levels", ["edge"])
        return stability
