# --------------------------------------------------------------------
# Copyright © 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

import json
import os
import sys
import subprocess

import argparse
from pkg_resources import resource_string as resource_bytes
from prettytable import PrettyTable, PLAIN_COLUMNS
from snappy.click import ClickDataSource
from snappy.systemimage import SystemImageDataSource
from snappy.utils import sudo

# FIXME: move to click
from click.build import ClickBuilder, ClickBuildError

__version__ = resource_bytes('snappy', 'version.txt').decode('utf-8').strip()


COMMASPACE = ', '


# FIXME: once the store really supports no namespaces we can get rid of this
DEFAULT_NAMESPACE_PREFIX = 'com.ubuntu.snappy.'
DEFAULT_NAMESPACE_TEMPLATE = DEFAULT_NAMESPACE_PREFIX + '{}'


def add_namespace_prefix(package):
    '''
    Convert a short package name to fully-qualified package name and return it
    '''
    if "." in package:
        return package
    if not package.startswith(DEFAULT_NAMESPACE_PREFIX):
        return DEFAULT_NAMESPACE_TEMPLATE.format(package)
    return package


def remove_namespace_prefix(package):
    '''
    Convert a fully-qualified package name and return the
    short name for it.
    '''
    if not package.startswith(DEFAULT_NAMESPACE_PREFIX):
        return package
    return package[len(DEFAULT_NAMESPACE_PREFIX):]


class MyHelpFormatter(argparse.HelpFormatter):
    # LP: #1397915
    def add_usage(self, usage, actions, groups, prefix=""):
        """Ensure that 'Usage:' is upper-case"""
        return super(MyHelpFormatter, self).add_usage(
            usage, actions, groups, prefix="Usage:")


class Main:
    def __init__(self):
        self.parser = argparse.ArgumentParser(
            prog='snappy',
            description='snappy command line interface',
            formatter_class=MyHelpFormatter)
        self.parser.add_argument(
            '-v', '--version',
            action='version', version=__version__,
            help='Print this version string and exit')
        commands = self.parser.add_subparsers(title='Commands')
        self.add_subcommand(commands, 'info', self._do_info)
        versions = self.add_subcommand(
            commands, 'versions', self._do_versions)
        versions.add_argument(
            '-a', '--all', action='store_true', default=False,
            help='show all versions')
        versions.add_argument(
            'args', nargs='*',
            help='Show specific versions')

        search = self.add_subcommand(commands, 'search', self._do_search)
        search.add_argument(
            'args', nargs='+',
            help='Search for a snappy package')

        self.add_subcommand(
            commands, 'update-versions', self._do_update_versions)

        update = self.add_subcommand(commands, 'update', self._do_update)
        update.add_argument(
            '--list', action='store_true',
            help='Only list what would be done')
        update.add_argument(
            '--user', action='store',
            help='Update package for USER')
        update.add_argument(
            'args', nargs='*',
            help='Additional optional arguments')

        rollback = self.add_subcommand(commands, 'rollback',
                                       self._do_rollback,
                                       help='''
                                       undo last system-image update.
                                       ''')
        rollback.add_argument(
            'args', nargs='+',
            help='Part to roll back')

        install = self.add_subcommand(commands, 'install', self._do_install)
        install.add_argument(
            '--user', action='store',
            help='Install package for USER')
        install.add_argument(
            '--stability', action='store',
            help='Requested stability level')
        install.add_argument(
            'args', nargs='*',
            help='Additional optional arguments')

        uninstall = self.add_subcommand(
            commands, 'uninstall', self._do_uninstall)
        uninstall.add_argument(
            '--user', action='store',
            help='Uninstall an installed package for USER')
        uninstall.add_argument(
            'args', nargs='*',
            help='Additional optional arguments')

        tags = self.add_subcommand(commands, 'tags', self._do_tags)
        tags.add_argument(
            'parts', nargs=1,
            help='part argument')

        build = self.add_subcommand(commands, 'build', self._do_build)
        build.add_argument(
            'directory', nargs=1,
            help='directory')

        # click stuff
        self.add_passthrough_subcommand(commands, 'chroot')
        self.add_passthrough_subcommand(commands, 'framework')

        # special, hidden
        fake_version = commands.add_parser(
            'fake-version', help=argparse.SUPPRESS)
        fake_version.set_defaults(func=self._do_fake_version)
        fake_version.add_argument('part', nargs=1)
        fake_version.add_argument('version', nargs=1)

        nap = commands.add_parser(
            'nap', help=argparse.SUPPRESS)
        nap.set_defaults(func=lambda a: sys.stdout.write("zzzZZZ\n"))

    def __main__(self):
        args = self.parser.parse_args()
        callback = getattr(args, 'func', None)
        if callback is None:
            self.parser.print_usage(sys.stderr)
            return 1
        return callback(args)

    def add_subcommand(self, parser, name, func, help=None):
        subparser = parser.add_parser(name, help=help)
        subparser.set_defaults(func=func)
        return subparser

    def add_passthrough_subcommand(self, parser, name):
        subparser = self.add_subcommand(
            parser,
            name,
            lambda *args: self._do_passthrough(name, *args))
        subparser.add_argument(
            'args', nargs='*',
            help='Additional optional arguments')
        return subparser

    def _do_fake_version(self, args):
        part = args.part[0]
        version = args.version[0]
        ret = subprocess.call(["mount", "-o", "remount,rw", "/"])
        if ret != 0:
            raise Exception("Failed to remount / rw")
        json_str = subprocess.check_output(["click", "info", part])
        click_dir = json.loads(json_str.decode("utf-8"))["_directory"]
        mpath = os.path.join(click_dir, ".click", "info", part+".manifest")
        with open(mpath) as fp:
            manifest = json.load(fp)
        manifest["version"] = version
        with open(mpath, "w") as fp:
            json.dump(manifest, fp, ensure_ascii=False, sort_keys=True,
                      indent=4, separators=(",", ": "))
        ret = subprocess.call(["mount", "-o", "remount,ro", "/"])
        return 0

    def _do_passthrough(self, command, args):
        command = ['click', command] + args.args
        return subprocess.call(command)

    def _do_build(self, args):
        builder = ClickBuilder()
        builder.add_file(args.directory[0], "./")
        try:
            path = builder.build(".", manifest_path="meta/package.yaml")
            os.rename(path, path.replace(".click", ".snap"))
        except ClickBuildError as e:
            print(e, file=sys.stderr)
            return 1

    def _do_tags(self, args):
        if args.parts == ["ubuntu-core"]:
            print(",".join(SystemImageDataSource().get_tags(args.parts)))
        else:
            print(",".join(ClickDataSource().get_tags(args.parts[0])))
        return 0

    def _do_info(self, args):
        print("release: {}".format(SystemImageDataSource().channel))
        print("frameworks: {}".format(
            ", ".join(ClickDataSource().installed_frameworks)))
        print("apps: {}".format(
            ", ".join(ClickDataSource().installed_apps)))
        return 0

    def _do_update_versions(self, args):
        updates = 0
        for part in (SystemImageDataSource().versions() +
                     ClickDataSource().versions()):
            # FIXME: use version_compare
            if (part.latest_version and
                    part.current_version < part.latest_version):
                updates += 1
        print("%s updated components are available with your current "
              "stability settings." % updates)

    def _do_versions(self, args, force_all=False):
        def display_active(part, no_reboot, newest):
            if part.active:
                return '*'
            elif part.installed and part.name == 'ubuntu-core':
                if no_reboot or part.current_version < newest:
                    return '-'
                else:
                    # Needs a reboot to make active
                    return "R"
            else:
                return '-'
        t = PrettyTable(["Part", "Tag", "Installed", "Available",
                         "Fingerprint", "Active"])
        t.set_style(PLAIN_COLUMNS)
        t.right_padding_width = 2
        t.align["Part"] = "l"
        t.align["Tag"] = "l"
        t.align["Installed"] = "l"
        t.align["Available"] = "l"
        t.align["Fingerprint"] = "l"
        t.align["Active"] = "l"
        all = getattr(args, "all", False) or force_all

        reboot_part = None
        current_version = -1
        other_version = -1

        parts = (SystemImageDataSource().versions(all) +
                 ClickDataSource().versions(all))

        # if True, do not flag a reboot as being required
        no_reboot = False

        # Check if both the current and other partitions have the same
        # system image version installed (as happens immediately after
        # provisioning). If so, no reboot is required.
        latest_system_versions = {}
        for part in parts:
            if part.name == 'ubuntu-core':
                latest_system_versions[int(part.current_version)] = \
                    part.current_hash

                if part.active:
                    current_version = int(part.current_version)
                elif part.installed:
                    other_version = int(part.current_version)
                    reboot_part = part.name

        if len(latest_system_versions) == 1:
            # Both A and B partitions have the same system-image
            # version installed, so don't want that a reboot is required
            no_reboot = True

        if other_version == -1:
            newest = current_version
        else:
            newest = max(current_version, other_version)

        for part in parts:

            if not args.args or part.name in args.args:
                t.add_row(
                    (part.name,
                     part.tag,
                     part.current_version,
                     part.latest_version or "-",
                     part.current_hash[:14],
                     display_active(part, no_reboot, newest)))

        print(t.get_string(padding_width=1))

        if (other_version > current_version) and reboot_part:

            # FIXME: we can't currently determine whether a 'snappy update'
            # has already been run.
            #
            # Once we introduce the /boot/grub partition we can have
            # 'snappy update' set a value using grub-editenv to store
            # the full path to the rootfs that will be booted *next*.
            # We can then read back that value here to determine the
            # correct message.
            sys.stdout.write("Reboot to use the new ubuntu-core.\n")
            sys.stdout.flush()
        return 0

    def _do_search(self, args):
        results = ClickDataSource().search(args.args)
        # No error occurred.  If the output is empty that means nothing was
        # found, otherwise just print the results from `click search`.
        if not results:
            print('No matching packages found: {}'.format(
                COMMASPACE.join(args.args)))
            return

        t = PrettyTable(["Part", "Version", "Description"])
        t.set_style(PLAIN_COLUMNS)
        t.right_padding_width = 2
        t.align["Part"] = "l"
        t.align["Version"] = "l"
        t.align["Description"] = "l"

        for result in results:
            # display the short name only
            part = remove_namespace_prefix(result["name"])
            ver = result.get("version", "")
            desc = result["title"]

            t.add_row((part, ver, desc))

        # remove extraneous right padding
        table = t.get_string(padding_width=1)
        table = [row.rstrip() for row in table.splitlines()]
        print("\n".join(table))

    def _do_install(self, args):
        if len(args.args) == 0:
            # There must be at least one argument, i.e. the package name.
            self.parser.print_usage(sys.stderr)

        packages = []
        for package in args.args:
            if os.path.exists(package):
                packages.append(package)
                continue

            # Add a shortcut convenience.
            if not os.path.exists(package):
                package = add_namespace_prefix(package)

            # check if its actually available
            manifest = ClickDataSource().details(package)
            if manifest is None:
                channel = SystemImageDataSource().channel
                print("No package '%s' for %s" % (
                    remove_namespace_prefix(package), channel))
                return 1
            # name may have changed, so update it
            packages.append(package)

        command = sudo() + ['click', 'install', '--meter'] \
                         + packages + [package]
        if args.user is None:
            command.append('--all-users')
        else:
            command.extend(('--user', args.user))
        # FIXME: *only* do this if the image is in developer mode.
        if os.path.exists(package):
            command.append('--allow-unauthenticated')
        # Ugly passthru to click(1) for now.
        ret_code = subprocess.call(command)
        # show versions
        if ret_code == 0:
            self._do_versions(args)
        return ret_code

    def _do_uninstall(self, args):
        if len(args.args) == 0:
            # There must be at least one argument, i.e. the package name.
            self.parser.print_usage(sys.stderr)
        # Add a shortcut convenience.
        package = args.args.pop(-1)
        command = sudo() + ['click', 'unregister'] + args.args + [package]
        if args.user is None:
            command.append('--all-users')
        else:
            command.extend(('--user', args.user))
        # Ugly passthru to click(1) for now.
        return subprocess.call(command)

    def _do_update(self, args):
        SystemImageDataSource().update(args.args, list_only=args.list)
        # yay, namespace handling :(((
        if not args.args:
            args.args = [remove_namespace_prefix(pkg)
                         for pkg in ClickDataSource().upgradable_apps]
            # nothing to upgrade
            if not args.args:
                return 0
        return self._do_install(args)

    def _do_rollback(self, args):
        res = SystemImageDataSource().rollback(args.args)
        res &= ClickDataSource().rollback(args.args)
        if res:
            return 0
        else:
            return 1


def main():
    return Main().__main__()
