#!/usr/bin/python3
# --------------------------------------------------------------------
# Copyright © 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

import os
import sys
import re
import logging
import logging.handlers
import atexit
import subprocess
import signal

from snappy.utils import sudo

from enum import Enum

log = logging.getLogger()


def shutdown_logging():
    logging.shutdown()


def setup_logger(debug=False):

    atexit.register(shutdown_logging)

    if debug:
        log.setLevel(logging.DEBUG)

    # send data to syslog...
    handler = logging.handlers.SysLogHandler(address='/dev/log')

    # add an identifier
    script = os.path.splitext(os.path.basename(__file__))[0]
    handler.ident = '{}: '.format(script)

    log.addHandler(handler)

    # ... and stderr
    handler = logging.StreamHandler()
    log.addHandler(handler)


class Partition():
    '''
    Utilities to work with snappy disk partitions.
    '''
    # Name of writable user data partition label as created by
    # ubuntu-device-flash(1).
    WRITABLE_PARTITION_LABEL = 'writable'

    # name of primary root filesystem partition label as created by
    # ubuntu-device-flash(1).
    ROOTFS_A_LABEL = 'system-a'

    # name of primary root filesystem partition label as created by
    # ubuntu-device-flash(1). Note that this partition will
    # only be present if this is an A/B upgrade system.
    ROOTFS_B_LABEL = 'system-b'

    ROOT_PARTITIONS = (ROOTFS_A_LABEL, ROOTFS_B_LABEL)

    # name of boot partition label as created by ubuntu-device-flash(1).
    BOOT_PARTITION_LABEL = 'boot'

    # FIXME: Should query system-image-cli (see bug LP:#1380574).
    DEFAULT_CACHE_DIR = '/writable/cache'

    # Directory to mount writable root filesystem below the cache
    # diretory.
    MOUNT_TARGET = 'system'

    system_types = Enum('SYSTEM_TYPE',
                        'SINGLE_ROOT ' +
                        'DUAL_ROOT ')

    # File creation mode used when any directories are created.
    DIR_MODE = 0o750

    # name of system-image's master configuration file. Used to query
    # the system-image version on the other partition.
    SYSTEM_IMAGE_CONFIG = '/etc/system-image/client.ini'

    # Grub's requirements (for working inside a chroot)
    required_filesystems = ('/dev', '/proc', '/sys')

    def __init__(self):
        if os.environ.get('SNAPPY_DEBUG'):
            setup_logger(debug=True)

        signal.signal(signal.SIGINT, self.signal_handler)

        # list of current mounts that this module has created
        self.mounts = []

        # list of current bindmounts this module has created
        self.bindmounts = []

        # all partitions
        self.partitions = self.get_partitions()

        # just root partitions
        self.roots = self.get_root_partitions()

        # should the rootfs be mounted read-only?
        self.read_only = False

        if self.uses_ab_partitions():
            self.system_type = self.system_types.DUAL_ROOT
        else:
            self.system_type = self.system_types.SINGLE_ROOT

        self.mount_target = self.get_mount_target()

    def signal_handler(self, signum, frame):
        '''
        Undo mount operations and exit on interrupt.
        '''
        if signum == signal.SIGINT:
            for mount in reversed(self.mounts):
                self.unmount(mount)
        sys.exit(0)

    def update_bootloader(self):
        '''
        Update the bootloader configuration for the system to ensure
        that the next boot will use the correct root filesystem
        partition.
        '''
        self.toggle_bootloader_rootfs()

    def get_cache_dir(self):
        '''
        Returns the full path to the cache directory, which is used as a
        scratch pad, for downloading new images to and bind mounting the
        rootfs.
        '''
        return self.DEFAULT_CACHE_DIR

    def get_mount_target(self):
        '''
        Get the full path to the mount target directory.
        '''
        return os.path.join(self.get_cache_dir(), self.MOUNT_TARGET)

    def get_partitions(self):
        '''
        Returns a list of tuples of the recognised root filesystem partitions
        available on this system. The tuples contain the following triplet:

        ( <partition-name>, <full-device-path>, <mountpoint>, <parent-name> )

        '''
        cmd = 'lsblk'

        # list of all the partitions we care about
        recognised = []

        recognised.extend(self.ROOT_PARTITIONS)
        recognised.append(self.BOOT_PARTITION_LABEL)
        recognised.append(self.WRITABLE_PARTITION_LABEL)

        matches = []

        args = [cmd, '--ascii', '--output-all', '--pairs']
        log.debug('running: {}'.format(args))

        proc = subprocess.Popen(sudo() + args,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.DEVNULL,
                                universal_newlines=True)

        if proc.wait() != 0:
            return matches

        stdout = proc.communicate()[0]
        lines = stdout.split('\n')

        for line in lines:
            line = line.rstrip()
            if not line:
                continue

            fields = {}

            # split the line into 'NAME="quoted value"' fields
            results = re.findall(r'(?:[^\s"]|"(?:[^"])*")+', line)
            for result in results:
                name, value = result.split('=')

                # remove quotes
                value = value.lstrip('"').rstrip('"')

                fields[name] = value

            if fields['NAME'] != fields['KNAME']:
                # For SCSI devices, these fields match
                continue

            if 'LABEL' in fields and fields['LABEL'] in recognised:
                # reconstructing the device name like this is valid
                # for SCSI devices
                device = '/dev/{}'.format(fields['NAME'])

                if not os.path.exists(device):
                    continue
                matches.append((fields['LABEL'],
                                device,
                                fields['MOUNTPOINT'],
                                fields['PKNAME']))

        return matches

    def get_root_partitions(self):
        '''
        Returns a list of tuples of the recognised root filesystem partitions
        available on this system. The tuples contain the following triplet:

        ( <partition-name>, <full-device-path>, <mountpoint>, <parent-name> )


        Requires that get_partitions() has already been called.
        '''
        assert(self.partitions)

        roots = [part for part in self.partitions
                 if part[0] in self.ROOT_PARTITIONS]
        return roots

    def get_boot_partition(self):
        '''
        Returns a tuple for the boot partition containing the following
        triplet (or None if no appropriate partition found):

        ( <partition-name>, <full-device-path>, <mountpoint>, <parent-name> )

        '''
        assert(self.partitions)

        for part in self.partitions:
            if part[0] == self.BOOT_PARTITION_LABEL:
                return part

        return None

    def get_writable_partition(self):
        '''
        Returns a tuple for the writable partition containing the following
        triplet (or None if no appropriate partition found):

        ( <partition-name>, <full-device-path>, <mountpoint>, <parent-name> )

        '''
        assert(self.partitions)

        for part in self.partitions:
            if part[0] == self.WRITABLE_PARTITION_LABEL:
                return part

        return None

    def uses_ab_partitions(self):
        '''
        Returns: True if the system uses A/B partitions, else False.
        '''
        return len(self.roots) == 2

    def get_rootfs(self):
        '''
        Returns the full path to the currently booted root partition.
        '''
        if self.roots[0][2] == '/':
            return self.roots[0][1]
        else:
            return self.roots[1][1]

    def get_other_rootfs(self):
        '''
        Returns the full path to the "other" partition for
        systems that have dual root filesystems (A/B partitions)
        '''
        if self.roots[0][2] == '/':
            return self.roots[1][1]
        else:
            return self.roots[0][1]

    def fsck(self, device):
        '''
        Run fsck(8) on specified device.

        Returns: True on success, else False.
        '''
        assert (os.path.exists(device))

        failed = False

        cmd = '/sbin/fsck'
        args = []

        args.append(cmd)

        # Paranoia - don't fsck if already mounted
        args.append('-M')

        args.append('-av')
        args.append(device)

        log.debug('running: {}'.format(args))

        proc = subprocess.Popen(sudo() + args,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                universal_newlines=True)

        ret = proc.wait()

        if ret == 0:
            return True

        stdout, stderr = proc.communicate()

        # fsck's return code 1 means: "Filesystem errors corrected"
        # (aka a warning - FS is consistent [now]).
        failed = False if ret == 1 else True

        log.error('{} returned {} ({}): {}, {}'
                  .format(args,
                          ret,
                          "failed" if failed else "warning",
                          stdout,
                          stderr))

        return failed

    def mount(self, source, target, options=None):
        '''
        Mount @source on @target using optional specified
        options string (which must not contain spaces).
        '''

        cmd = 'mount'

        args = []

        args.append(cmd)
        if options:
            args += ['-o', options]
        args.append(source)
        args.append(target)

        log.debug('running: {}'.format(args))

        proc = subprocess.Popen(sudo() + args,
                                stderr=subprocess.PIPE,
                                universal_newlines=True)
        if proc.wait() != 0:
            stderr = proc.communicate()[1]
            log.error('failed to mount ({}): {}'
                      .format(args, stderr))
            sys.exit(1)

        self.mounts.append(target)

    def bind_mount(self, source, target):
        '''
        Bind mount @source to @target.
        '''
        log.debug('bind mounting existing root')
        self.mount(source, target, 'bind')

    def make_mount_point(self):
        '''
        Ensure the system mount point exists.
        '''
        cmd = 'mkdir'
        args = []
        args.append(cmd)
        args.append('-p')

        if self.read_only:
            # A system update "owns" the default mount_target directory.
            # So in read-only mode, use a temporary mountpoint name to
            # avoid colliding with a running system update.
            self.mount_target = '{}-ro-{}'.format(self.mount_target,
                                                  os.getpid())

        args.append(self.mount_target)

        log.debug('running: {}'.format(args))

        proc = subprocess.Popen(sudo() + args,
                                stdout=subprocess.DEVNULL,
                                stderr=subprocess.DEVNULL,
                                universal_newlines=True)

        if proc.wait() != 0:
            log.error('failed to create directory {}:'
                      .format(self.mount_target))
            sys.exit(1)

    def mount_other_rootfs(self, read_only=False):
        '''
        Mount the other root filesystem.
        '''
        self.read_only = read_only

        self.make_mount_point()

        self.root = self.get_other_rootfs()

        if not self.read_only:
            # needs to be mounted writable, so check it first!
            self.fsck(self.root)
            self.mount(self.root, self.mount_target)
        else:
            self.mount(self.root, self.mount_target, options='ro')

    def unmount_other_rootfs(self):
        '''
        Unmount the root filesystem for the partition that is
        to be upgraded.
        '''
        self.unmount(self.mount_target)

    def unmount(self, target, options=None):
        '''
        Unmount the specified mount target, using the specified list of
        options.

        Errors are fatal.
        '''

        log.debug('unmounting {}'.format(target))

        args = []
        args.append('umount')
        if options:
            args.extend(options)
        args.append(target)

        log.debug('running: {}'.format(args))

        proc = subprocess.Popen(sudo() + args,
                                stderr=subprocess.PIPE,
                                universal_newlines=True)

        if proc.wait() != 0:
            stderr = proc.communicate()[1]
            log.error('failed to unmount ({}): {}'
                      .format(args, stderr))
            sys.exit(1)

        self.mounts.remove(target)

    def root_rmdir(self, directory):
        '''
        Delete the specified directory by running as root user.
        '''
        cmd = 'rmdir'

        args = []
        args.append(cmd)
        args.append(directory)

        log.debug('running: {}'.format(args))

        proc = subprocess.Popen(sudo() + args,
                                stdout=subprocess.DEVNULL,
                                stderr=subprocess.DEVNULL,
                                universal_newlines=True)

        if proc.wait() != 0:
            log.debug('Failed to remove directory: {}'
                      .format(directory))

    def get_other_version(self):
        '''
        Returns a tuple containing the following:

            (<other-version>, <other-version-detail>)

        Where,

        - <other-version> is the numeric system-image version
          installed on the other partition, or -1 on error.
        - <other-version-detail> is the version_detail field from
          system-image-cli(1), or None on error.

        '''
        self.mount_other_rootfs(read_only=True)

        # full path to other partitions system-image master config file
        file = os.path.normpath('{}/{}'
                                .format(self.mount_target,
                                        self.SYSTEM_IMAGE_CONFIG))

        if not os.path.exists(file):
            self.unmount(self.mount_target)
            self.root_rmdir(self.mount_target)
            return (-1, None)

        # query the version of the s-i on the other partition.
        cmd = 'system-image-cli'
        args = []

        args.append(cmd)
        args.append('-C')
        args.append(file)
        args.append('--info')

        log.debug('running: {}'.format(args))

        proc = subprocess.Popen(sudo() + args,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.DEVNULL,
                                universal_newlines=True)
        if proc.wait() != 0:
            self.unmount(self.mount_target)
            self.root_rmdir(self.mount_target)
            return (-1, None)

        stdout = proc.communicate()[0]
        lines = stdout.split('\n')

        version_pattern = 'version version: (\d+)'
        version_re = re.compile(version_pattern)
        other_version = None

        for line in lines:
            line = line.rstrip()

            result = version_re.match(line)
            if result:
                other_version = int(result.group(1))
                break

        if other_version:
            self.unmount(self.mount_target)
            self.root_rmdir(self.mount_target)
            details = self._get_version_detail()
            return (other_version, details)

        self.unmount(self.mount_target)
        self.root_rmdir(self.mount_target)

        return (-1, None)

    def _get_version_detail(self):
        '''
        Currently, system-image-cli(1) does not provide 'version_string'
        in its verbatim form, hence grope for it for now.

        Returns: details string, or None on error
        '''
        cmd = 'ubuntu-core-upgrade'
        args = []

        args.append(cmd)
        args.append('--show-other-details')

        log.debug('running: {}'.format(args))

        proc = subprocess.Popen(sudo() + args,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.DEVNULL,
                                universal_newlines=True)
        if proc.wait() != 0:
            return None

        detail_pattern = 'version_detail: (.*)'
        detail_re = re.compile(detail_pattern)

        stdout = proc.communicate()[0]
        lines = stdout.split('\n')

        for line in lines:
            line = line.rstrip()

            result = detail_re.match(line)
            if result:
                return result.group(1)

        return None

    def bindmount_required_filesystems(self):
        '''
        Grub requires a few virtual filesystems to be mounted
        for it to work.
        '''
        for fs in self.required_filesystems:
            target = os.path.normpath('{}/{}'.format(self.mount_target, fs))
            self.bind_mount(fs, target)
            self.bindmounts.append(target)

        boot = self.get_boot_partition()
        if not boot:
            # No separate boot partition
            return

        mountpoint = boot[2]
        if not mountpoint:
            # Should never happen - partition exists but not already
            # mounted in main FS context.
            return

        target = os.path.normpath('{}/{}'.format(self.mount_target,
                                                 mountpoint))

        self.bind_mount(mountpoint, target)
        self.bindmounts.append(target)

    def unmount_required_filesystems(self):
        '''
        Undo bindmount_required_filesystems().
        '''
        for target in self.bindmounts:
            self.unmount(target)

    def install_grub(self):

        args = []
        args.append('grub-install')

        device = None

        # Find the name of the underlying disk
        for root in self.roots:
            if root[1] == self.root:
                device = '/dev/{}'.format(root[3])
                break

        if not device or not os.path.exists(device):
            sys.exit('ERROR: cannot determine device for root {}'
                     .format(self.root))

        args.append(device)

        # install grub from the other parition
        self.run_in_chroot(args)

        # create the grub config
        args = ['update-grub']

        self.run_in_chroot(args)

    def toggle_bootloader_rootfs(self):
        '''
        Called after the upgrader has run on the other partition to
        inform the bootloader to boot into the "other" partition,
        rather than the current one.
        '''

        self.mount_other_rootfs()
        self.bindmount_required_filesystems()

        self.install_grub()

        self.unmount_required_filesystems()
        self.unmount_other_rootfs()

    def run_in_chroot(self, args):
        '''
        Run the commandline specified by the args array chrooted to the
        new root filesystem.

        Errors are fatal.
        '''
        log.debug('running in chroot ({}): {}'
                  .format(self.mount_target, args))

        full_args = []
        full_args += sudo()
        full_args += ['/usr/sbin/chroot', self.mount_target]
        full_args += args

        log.debug('full command-line: {}'.format(full_args))

        proc = subprocess.Popen(full_args,
                                stdin=subprocess.DEVNULL,
                                stdout=subprocess.DEVNULL,
                                stderr=subprocess.DEVNULL,
                                universal_newlines=True)

        ret = proc.wait()
        if ret != 0:
            sys.exit('{} failed'.format(full_args))


def main():
    if os.getuid() != 0:
        sys.exit("ERROR: need to be root")

    setup_logger(debug=True)

    partition = Partition()

    other_version = partition.get_other_version()
    print('XXX: other partition running version: {}'
          .format(other_version))

    partition.update_bootloader()


if __name__ == '__main__':
    main()
