# --------------------------------------------------------------------
# Copyright © 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

import hashlib
import sys
import math

import dbus

from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)

from gi.repository import GObject
from progressbar import ProgressBar, Bar, Widget

from . import datasource
from snappy.partition import Partition


# XXX This should really be refactored with click/acquire.py's
# ClickAcquireStatusMeter.  They are essentially the same implementation, but
# there are some API differences in the caller which prevent them from being
# shared currently.  Cargo culting.

class FileSize(Widget):
    def update(self, pbar):
        value = pbar.maxval                         # or .currval?
        if value < 1000:
            return '{:3d} B '.format(int(value))
        scale = int(math.log(value, 1000))          # or binary? e.g. MiB
        try:
            suffix = ' kMGTPEZY'[scale]
        except IndexError:
            # Should never happen, but still.
            suffix = '!'
        scaled_value = value // 1000 ** scale
        return '{:3d} {:1s}B '.format(scaled_value, suffix)


class OK(Widget):
    def update(self, pbar):
        if pbar.currval >= pbar.maxval:
            return 'OK'
        else:
            return '  '

SPACER = '    '


class SystemImageDataSource(datasource.DataSource):

    core_part = 'ubuntu-core'

    def __init__(self):
        self.bus = dbus.SystemBus()
        self._obj = self.bus.get_object(
            "com.canonical.SystemImage", "/Service")
        self._iface = dbus.Interface(self._obj, "com.canonical.SystemImage")

        # Disable automatic downloading since it slows down simply
        # checking the latest server version available.
        self._iface.SetSetting("auto_download", "0")

        self._iface.connect_to_signal("UpdateAvailableStatus", self._update)
        self._iface.connect_to_signal("UpdateProgress", self._update_progress)
        self._iface.connect_to_signal("UpdateFailed", self._update_failed)
        self._iface.connect_to_signal(
            "UpdateDownloaded", self._update_downloaded)
        self._iface.connect_to_signal("Rebooting", self._update_applied)
        self.loop = GObject.MainLoop()
        # FIXME: what should be the default here?
        self._verbose = False
        self._list_only = False
        self._latest_version = "-"
        self._update_size = None
        self._progress = None

    def _update(self, is_available, downloading, available_version,
                update_size, last_update_data, error_reason, sender=None):
        if not is_available:
            # print("Your system is up-to-date")
            self.loop.quit()
            return
        # print("Update for the core-system is available", available_version)
        if self._list_only:
            self._latest_version = available_version
            self.loop.quit()
            return
        # Set up the progress meter.
        if self._progress is None:
            self._update_size = update_size
            self._progress = ProgressBar(widgets=[
                'ubuntu-core',
                SPACER,
                # XXX in the examples, it's unclear whether the size field
                # should be the total size or the fetched size.  Also, decimal
                # (MB) or binary (MiB)?  For now, we go with total size in
                # decimal.
                FileSize(),
                SPACER,
                Bar(marker='=', left='[', right=']'),
                SPACER,
                OK(),
                SPACER,
                ],
                maxval=self._update_size)
            self._progress.start()
        if not downloading:
            self._iface.DownloadUpdate()

    def _update_progress(self, percent, eta):
        # #if self._verbose:
        # #    sys.stdout.write("\rDownloading: %s" % percent)
        # #    sys.stdout.flush()
        received_bytes = self._update_size * percent // 100
        if received_bytes <= 1:
            return
        self._progress.update(received_bytes)

    def _update_failed(self, count, reason):
        print("failed", count, reason)
        self.loop.quit()

    def _update_downloaded(self):
        # #sys.stdout.write("\nInstalling: ")
        # #sys.stdout.flush()
        self._iface.ApplyUpdate()

    def _update_applied(self, status):
        '''
        Called after the system-image update has been applied.

        @status: True if a reboot is required, else False.
        '''
        self.loop.quit()

        if not status:
            # no reboot required
            return

        # If the update was fully applied and a reboot is therefore
        # required, update the bootloader to ensure the next boot uses
        # the other partition for the rootfs.

        partition = Partition()
        partition.update_bootloader()
        if self._progress is not None:
            self._progress.finish()
        sys.stdout.write("Reboot to use the new ubuntu-core.\n")
        sys.stdout.flush()

    @property
    def channel(self):
        info = self._iface.Information()
        return info["channel_name"]

    def versions(self, all_versions=False):
        saved = self._list_only
        self._list_only = True
        self._iface.CheckForUpdate()
        self.loop.run()
        self.info = self._iface.Information()
        if int(self.info["target_build_number"]) > 0:
            latest_version = self.info["target_build_number"]
        else:
            latest_version = ""
        if latest_version == self.info["current_build_number"]:
            latest_version = 0
        part = datasource.Part(
            self.core_part,
            self.info.get("current_stability", "edge"),
            self.info["current_build_number"],
            latest_version,
            # version_detail identifies the image fully, i.e. it contains
            # all part files (ubuntu,device,custom and it channel version)
            hashlib.sha256(self.info["version_detail"]
                           .encode("utf-8")).hexdigest()
        )
        parts = []
        parts.append(part)

        if all_versions:
            other_part = self.get_other_part()
            if other_part:
                parts.append(other_part)

        # reset to allow this method to be called by other methods
        # without changing the expected behaviour.
        self._list_only = saved

        return parts

    def update(self, names_to_update=None, list_only=False):
        if not names_to_update:
            names_to_update = [self.core_part]
        self._list_only = list_only
        # check if we actually need to update anything
        if self.core_part not in names_to_update:
            return True
        # remove "ubntu-core" from args (this part has consumed it)
        names_to_update.remove(self.core_part)

        # latest system-image revision installed on either rootfs.
        newest_installed = -1
        active_version = -1

        parts = self.versions(True)
        latest_version = int(self.info["target_build_number"])

        for part in parts:
            if part.active:
                active_version = int(part.current_version)
            newest_installed = max(newest_installed,
                                   int(part.current_version))

        if newest_installed == latest_version:
            if active_version != latest_version:
                # System has been rolled back so simply toggle
                # bootloader to use the newest image on next boot.
                partition = Partition()
                partition.update_bootloader()

            # Latest revision already locally available
            return True

        # and now do the update
        self._iface.CheckForUpdate()

        self.loop.run()
        # FIXME: --list-only is not implemented yet

        return True

    def get_other_part(self):
        '''
        Returns a new Part object containing details from the other
        root partition, or None if the other partition is empty.
        '''

        # FIXME: see LP: #1396000
        other_tags = 'edge'

        # not sure what we're supposed to say here
        other_latest_version = 0

        partition = Partition()

        # determine system-image version installed on other partition
        other_version, other_detail = partition.get_other_version()

        if other_version == -1:
            return None

        other_fingerprint = hashlib.sha256(
            other_detail.encode("utf-8")).hexdigest()

        part = datasource.Part(
            self.core_part,
            other_tags,
            other_version,
            other_latest_version,
            other_fingerprint,
            active=False
        )

        return part

    def rollback(self, names_to_update=None):
        if not names_to_update:
            names_to_update = [self.core_part]
        # check if we actually need to update anything
        if self.core_part not in names_to_update:
            return True
        # remove "ubuntu-core" from args (this part has consumed it)
        names_to_update.remove(self.core_part)

        parts = self.versions(True)
        core = parts[0]
        current_version = int(core.current_version)

        partition = Partition()

        if not partition.uses_ab_partitions():
            sys.stdout.write('Rollback not supported on '
                             'single root system\n')
            return False

        # determine version details on the other root partition
        other_part = self.get_other_part()

        if not other_part or int(other_part.current_version) < 0:
            sys.stderr.write("Run 'snappy update' first\n")
            return False

        if int(other_part.current_version) > current_version:
            sys.stdout.write("You cannot rollback any further; "
                             "run 'snappy update' to move to a "
                             "newer version\n")

            return False

        current_tags = ','.join(self.get_tags(self.core_part))
        current_fingerprint = core.current_hash[:14]
        current_meta = '{} {} {}'.format(current_tags,
                                         current_version,
                                         current_fingerprint)

        other_meta = '{} {} {}'.format(other_part.tag,
                                       other_part.current_version,
                                       other_part.current_hash[:14])

        sys.stdout.write('Rolling back {}: ({} -> {})\n'
                         .format(self.core_part, current_meta, other_meta))

        partition.update_bootloader()

        sys.stdout.write("Reboot to use the new ubuntu-core.\n")
        sys.stdout.flush()
        return True

    def get_tags(self, part):
        # see LP bug #1396000
        try:
            tags = self._iface.GetTags()
        except dbus.exceptions.DBusException:
            tags = ["edge"]
        return tags
